@@ -4,7 +4,7 @@ use 5.010001;
 
 package RT::Extension::REST2;
 
-our $VERSION = '1.07';
+our $VERSION = '1.09';
 our $REST_PATH = '/REST/2.0';
 
 use Plack::Builder;
@@ -69,6 +69,7 @@ see a response, typical of search results, like this:
        "total" : 1,
        "count" : 1,
        "page" : 1,
+       "pages" : 1,
        "per_page" : 20,
        "items" : [
           {
@@ -224,6 +225,177 @@ adding time worked can be automatically be recalculated).
 You may of course choose to ignore the C<ETag> header and not provide
 C<If-Match> in your requests; RT doesn't require its use.
 
+=head3 Replying/Commenting Tickets
+
+You can reply to or comment a ticket by C<POST>ing to C<_url> from the
+C<correspond> or C<comment> hyperlinks that were returned when fetching the
+ticket.
+
+    curl -X POST
+         -H "Content-Type: application/json"
+         -d '{
+              "Subject"    : "response",
+              "Content"    : "What is your <em>issue</em>?",
+              "ContentType": "text/html",
+              "TimeTaken"  : "1"
+            }'
+         -H 'Authorization: token XX_TOKEN_XX'
+            'XX_TICKET_URL_XX'/correspond
+
+Replying or commenting a ticket is quite similar to a ticket creation: you
+send a C<POST> request, with data encoded in C<JSON>. The difference lies in
+the properties of the JSON data object you can pass:
+
+=over 4
+
+=item C<Subject>
+
+The subject of your response/comment, optional
+
+=item C<Content>
+
+The content of your response/comment, mandatory unless there is a non empty
+C<Attachments> property to add at least one attachment to the ticket (see
+L<Add Attachments> section below).
+
+=item C<ContentType>
+
+The MIME content type of your response/comment, typically C<text/plain> or
+C</text/html>, mandatory unless there is a non empty C<Attachments> property
+to add at least one attachment to the ticket (see L<Add Attachments> section
+below).
+
+=item C<TimeTaken>
+
+The time, in minutes, you've taken to work on your response/comment, optional.
+
+=back
+
+=head3 Add Attachments
+
+You can attach any binary or text file to your response or comment by
+specifying C<Attachements> property in the JSON object, which should be a
+JSON array where each item represents a file you want to attach. Each item
+is a JSON object with the following properties:
+
+=over 4
+
+=item C<FileName>
+
+The name of the file to attach to your response/comment, mandatory.
+
+=item C<FileType>
+
+The MIME type of the file to attach to your response/comment, mandatory.
+
+=item C<FileContent>
+
+The content, I<encoded in C<MIME Base64>> of the file to attach to your
+response/comment, mandatory.
+
+=back
+
+The reason why you should encode the content of any file to C<MIME Base64>
+is that a JSON string value should be a sequence of zero or more Unicode
+characters. C<MIME Base64> is a binary-to-text encoding scheme widely used
+(for eg. by web browser) to send binary data when text data is required.
+Most popular language have C<MIME Base64> libraries that you can use to
+encode the content of your attached files (see L<MIME::Base64> for C<Perl>).
+Note that even text files should be C<MIME Base64> encoded to be passed in
+the C<FileContent> property.
+
+Here's a Perl example to send an image and a plain text file attached to a
+comment:
+
+    #!/usr/bin/perl
+    use strict;
+    use warnings;
+
+    use LWP::UserAgent;
+    use JSON;
+    use MIME::Base64;
+    use Data::Dumper;
+
+    my $url = 'http://rt.local/REST/2.0/ticket/1/comment';
+
+    my $img_path = '/tmp/my_image.png';
+    my $img_content;
+    open my $img_fh, '<', $img_path or die "Cannot read $img_path: $!\n";
+    {
+        local $/;
+        $img_content = <$img_fh>;
+    }
+    close $img_fh;
+    $img_content = MIME::Base64::encode_base64($img_content);
+
+    my $txt_path = '~/.bashrc';
+    my $txt_content;
+    open my $txt_fh, '<', glob($txt_path) or die "Cannot read $txt_path: $!\n";
+    {
+        local $/;
+        $txt_content = <$txt_fh>;
+    }
+    close $txt_fh;
+    $txt_content = MIME::Base64::encode_base64($txt_content);
+
+    my $json = JSON->new->utf8;
+    my $payload = {
+        Content => '<p>I want <b>two</b> <em>attachments</em></p>',
+        ContentType => 'text/html',
+        Subject => 'Attachments in JSON Array',
+        Attachments => [
+            {
+                FileName => 'my_image.png',
+                FileType => 'image/png',
+                FileContent => $img_content,
+            },
+            {
+                FileName => '.bashrc',
+                FileType => 'text/plain',
+                FileContent => $txt_content,
+            },
+        ],
+    };
+
+    my $req = HTTP::Request->new(POST => $url);
+    $req->header('Authorization' => 'token 6-66-66666666666666666666666666666666');
+    $req->header('Content-Type'  => 'application/json' );
+    $req->header('Accept'        => 'application/json' );
+    $req->content($json->encode($payload));
+
+    my $ua = LWP::UserAgent->new;
+    my $res = $ua->request($req);
+    print Dumper($json->decode($res->content)) . "\n";
+
+Encoding the content of attachments file in C<MIME Base64> has the drawback
+of adding some processing overhead and to increase the sent data size by
+around 33%. RT's REST2 API provides another way to attach any binary or text
+file to your response or comment by C<POST>ing, instead of a JSON request, a
+C<multipart/form-data> request. This kind of request is similar to what the
+browser sends when you add attachments in RT's reply or comment form. As its
+name suggests, a C<multipart/form-data> request message contains a series of
+parts, each representing a form field. To reply to or comment a ticket, the
+request has to include a field named C<JSON>, which, as previously, is a
+JSON object with C<Subject>, C<Content>, C<ContentType>, C<TimeTaken>
+properties. Files can then be attached by specifying a field named
+C<Attachments> for each of them, with the content of the file as value and
+the appropriate MIME type.
+
+The curl invocation is quite straightforward:
+
+    curl -X POST
+         -H "Content-Type: multipart/form-data"
+         -F 'JSON={
+                    "Subject"    : "Attachments in multipart/form-data",
+                    "Content"    : "<p>I want <b>two</b> <em>attachments</em></p>",
+                    "ContentType": "text/html",
+                    "TimeTaken"  : "1"
+                  };type=application/json'
+         -F 'Attachments=@/tmp/my_image.png;type=image/png'
+         -F 'Attachments=@/tmp/.bashrc;type=text/plain'
+         -H 'Authorization: token XX_TOKEN_XX'
+            'XX_TICKET_URL_XX'/comment
+
 =head3 Summary
 
 RT's REST2 API provides the tools you need to build robust and dynamic
@@ -311,11 +483,22 @@ Below are some examples using the endpoints above.
         -d '{ "Content": "Testing a correspondence", "ContentType": "text/plain" }'
         'https://myrt.com/REST/2.0/ticket/6/correspond'
 
-    # Comment a ticket
+    # Correspond a ticket with a transaction custom field
+    curl -X POST -H "Content-Type: application/json" -u 'root:password'
+        -d '{ "Content": "Testing a correspondence", "ContentType": "text/plain",
+              "TxnCustomFields": {"MyField": "custom field value"} }'
+        'https://myrt.com/REST/2.0/ticket/6/correspond'
+
+    # Comment on a ticket
     curl -X POST -H "Content-Type: text/plain" -u 'root:password'
         -d 'Testing a comment'
         'https://myrt.com/REST/2.0/ticket/6/comment'
 
+    # Comment on a ticket with custom field update
+    curl -X POST -H "Content-Type: text/plain" -u 'root:password'
+        -d '{ "Content": "Testing a comment", "ContentType": "text/plain", "CustomFields": {"Severity": "High"} }'
+        'https://myrt.com/REST/2.0/ticket/6/comment'
+
     # Create an Asset
     curl -X POST -H "Content-Type: application/json" -u 'root:password'
         -d '{"Name" : "Asset From Rest", "Catalog" : "General assets", "Content" : "Some content"}'
@@ -356,6 +539,11 @@ Below are some examples using the endpoints above.
     GET /attachment/:id
         retrieve an attachment
 
+=head3 Image and Binary Object Custom Field Values
+
+    GET /download/cf/:id
+        retrieve an image or a binary file as an object custom field value
+
 =head3 Queues
 
     GET /queues/all
@@ -528,8 +716,50 @@ Below are some examples using the endpoints above.
     POST /customfields
         search for custom fields using L</JSON searches> syntax
 
+    POST /customfield
+        create a customfield; provide JSON content
+
+    GET /catalog/:id/customfields?query=<JSON>
+    POST /catalog/:id/customfields
+        search for custom fields attached to a catalog using L</JSON searches> syntax
+
+    GET /class/:id/customfields?query=<JSON>
+    POST /class/:id/customfields
+        search for custom fields attached to a class using L</JSON searches> syntax
+
+    GET /queue/:id/customfields?query=<JSON>
+    POST /queue/:id/customfields
+        search for custom fields attached to a queue using L</JSON searches> syntax
+
     GET /customfield/:id
-        retrieve a custom field
+        retrieve a custom field, with values if type is Select
+
+    GET /customfield/:id?category=<category name>
+        retrieve a custom field, with values filtered by category if type is Select
+
+    PUT /customfield/:id
+        update a custom field's metadata; provide JSON content
+
+    DELETE /customfield/:id
+        disable customfield
+
+=head3 Custom Field Values
+
+    GET /customfield/:id/values?query=<JSON>
+    POST /customfield/:id/values
+        search for values of a custom field  using L</JSON searches> syntax
+
+    POST /customfield/:id/value
+        add a value to a custom field; provide JSON content
+
+    GET /customfield/:id/value/:id
+        retrieve a value of a custom field
+
+    PUT /customfield/:id/value/:id
+        update a value of a custom field; provide JSON content
+
+    DELETE /customfield/:id/value/:id
+        remove a value from a custom field
 
 =head3 Custom Roles
 
@@ -588,6 +818,36 @@ C<orderby=Created&order=ASC&orderby=id&order=DESC>. C<orderby> and
 C<order> query parameters are supported in both JSON and TicketSQL
 searches.
 
+The same C<field> is specified more than one time to express more than one
+condition on this field. For example:
+
+    [
+        { "field":    "id",
+          "operator": ">",
+          "value":    $min },
+
+        { "field":     "id",
+          "operator": "<",
+          "value":    $max }
+    ]
+
+By default, RT will aggregate these conditions with an C<OR>, except for
+when searching queues, where an C<AND> is applied. If you want to search for
+multiple conditions on the same field aggregated with an C<AND> (or an C<OR>
+for queues), you can specify C<entry_aggregator> keys in corresponding
+hashes:
+
+    [
+        { "field":    "id",
+          "operator": ">",
+          "value":    $min },
+
+        { "field":             "id",
+          "operator":         "<",
+          "value":            $max,
+          "entry_aggregator": "AND" }
+    ]
+
 Results are returned in
 L<the format described below|/"Example of plural resources (collections)">.
 
@@ -599,7 +859,9 @@ standard JSON format:
     {
        "count" : 20,
        "page" : 1,
+       "pages" : 191,
        "per_page" : 20,
+       "next_page" : "<collection path>?page=2"
        "total" : 3810,
        "items" : [
           { … },
@@ -611,12 +873,186 @@ standard JSON format:
 Each item is nearly the same representation used when an individual resource
 is requested.
 
+=head2 Object Custom Field Values
+
+When creating (via C<POST>) or updating (via C<PUT>) a resource which has
+some custom fields attached to, you can specify the value(s) for these
+customfields in the C<CustomFields> property of the JSON object parameter.
+The C<CustomFields> property should be a JSON object, with each property
+being the custom field identifier or name. If the custom field can have only
+one value, you just have to speciy the value as JSON string for this custom
+field. If the customfield can have several value, you have to specify a JSON
+array of each value you want for this custom field.
+
+    "CustomFields": {
+        "XX_SINGLE_CF_ID_XX"   : "My Single Value",
+        "XX_MULTI_VALUE_CF_ID": [
+            "My First Value",
+            "My Second Value"
+        ]
+    }
+
+Note that for a multi-value custom field, you have to specify all the values
+for this custom field. Therefore if the customfield for this resource
+already has some values, the existing values must be including in your
+update request if you want to keep them (and add some new values).
+Conversely, if you want to delete some existing values, do not include them
+in your update request (including only values you wan to keep). The
+following example deletes "My Second Value" from the previous example:
+
+    "CustomFields": {
+        "XX_MULTI_VALUE_CF_ID": [
+            "My First Value"
+        ]
+    }
+
+To delete a single-value custom field, set its value to JSON C<null>
+(C<undef> in Perl):
+
+    "CustomFields": {
+        "XX_SINGLE_CF_ID_XX" : null
+    }
+
+New values for Image and Binary custom fields can be set by specifying a
+JSON object as value for the custom field identifier or name with the
+following properties:
+
+=over 4
+
+=item C<FileName>
+
+The name of the file to attach, mandatory.
+
+=item C<FileType>
+
+The MIME type of the file to attach, mandatory.
+
+=item C<FileContent>
+
+The content, I<encoded in C<MIME Base64>> of the file to attach, mandatory.
+
+=back
+
+The reason why you should encode the content of the image or binary file to
+C<MIME Base64> is that a JSON string value should be a sequence of zero or
+more Unicode characters. C<MIME Base64> is a binary-to-text encoding scheme
+widely used (for eg. by web browser) to send binary data when text data is
+required. Most popular language have C<MIME Base64> libraries that you can
+use to encode the content of your attached files (see L<MIME::Base64> for
+C<Perl>). Note that even text files should be C<MIME Base64> encoded to be
+passed in the C<FileContent> property.
+
+    "CustomFields": {
+        "XX_SINGLE_IMAGE_OR_BINARY_CF_ID_XX"   : {
+            "FileName"   : "image.png",
+            "FileType"   : "image/png",
+            "FileContent": "XX_BASE_64_STRING_XX"
+        },
+        "XX_MULTI_VALUE_IMAGE_OR_BINARY_CF_ID": [
+            {
+                "FileName"   : "another_image.png",
+                "FileType"   : "image/png",
+                "FileContent": "XX_BASE_64_STRING_XX"
+            },
+            {
+                "FileName"   : "hello_world.txt",
+                "FileType"   : "text/plain",
+                "FileContent": "SGVsbG8gV29ybGQh"
+            }
+        ]
+    }
+
+Encoding the content of image or binary files in C<MIME Base64> has the
+drawback of adding some processing overhead and to increase the sent data
+size by around 33%. RT's REST2 API provides another way to upload image or
+binary files as custom field alues by sending, instead of a JSON request, a
+C<multipart/form-data> request. This kind of request is similar to what the
+browser sends when you upload a file in RT's ticket creation or update
+forms. As its name suggests, a C<multipart/form-data> request message
+contains a series of parts, each representing a form field. To create or
+update a ticket with image or binary file, the C<multipart/form-data>
+request has to include a field named C<JSON>, which, as previously, is a
+JSON object with C<Queue>, C<Subject>, C<Content>, C<ContentType>, etc.
+properties. But instead of specifying each custom field value as a JSON
+object with C<FileName>, C<FileType> and C<FileContent> properties, each
+custom field value should be a JSON object with C<UploadField>. You can
+choose anything you want for this field name, except I<Attachments>, which
+should be reserved for attaching files to a response or a comment to a
+ticket. Files can then be attached by specifying a field named as specified
+in the C<CustomFields> property for each of them, with the content of the
+file as value and the appropriate MIME type.
+
+Here is an exemple of a curl invocation, wrapped to multiple lines for
+readability, to create a ticket with a multipart/request to upload some
+image or binary files as custom fields values.
+
+    curl -X POST
+         -H "Content-Type: multipart/form-data"
+         -F 'JSON={
+                    "Queue"      : "General",
+                    "Subject"    : "hello world",
+                    "Content"    : "That <em>damned</em> printer is out of order <b>again</b>!",
+                    "ContentType": "text/html",
+                    "CustomFields"  : {
+                        "XX_SINGLE_IMAGE_OR_BINARY_CF_ID_XX"   => { "UploadField": "FILE_1",
+                        "XX_MULTI_VALUE_IMAGE_OR_BINARY_CF_ID" => [ { "UploadField": "FILE_2" }, { "UploadField": "FILE_3" } ]
+                    }
+                  };type=application/json'
+         -F 'FILE_1=@/tmp/image.png;type=image/png'
+         -F 'FILE_2=@/tmp/another_image.png;type=image/png'
+         -F 'FILE_3=@/etc/cups/cupsd.conf;type=text/plain'
+         -H 'Authorization: token XX_TOKEN_XX'
+            'XX_RT_URL_XX'/tickets
+
+If you want to delete some existing values from a multi-value image or
+binary custom field, you can just pass the existing filename as value for
+the custom field identifier or name, no need to upload again the content of
+the file. The following example will delete the text file and keep the image
+upload in previous example:
+
+    "CustomFields": {
+        "XX_MULTI_VALUE_IMAGE_OR_BINARY_CF_ID": [
+                "image.png"
+        ]
+    }
+
+To download an image or binary file which is the custom field value of a
+resource, you just have to make a C<GET> request to the entry point returned
+for the corresponding custom field when fetching this resource, and it will
+return the content of the file as an octet string:
+
+    curl -i -H 'Authorization: token XX_TOKEN_XX' 'XX_TICKET_URL_XX'
+
+    {
+        […]
+        "XX_IMAGE_OR_BINARY_CF_ID_XX" : [
+            {
+                "content_type" : "image/png",
+                "filename" : "image.png",
+                "_url" : "XX_RT_URL_XX/REST/2.0/download/cf/XX_IMAGE_OR_BINARY_OCFV_ID_XX"
+            }
+        ],
+        […]
+    },
+
+    curl -i -H 'Authorization: token XX_TOKEN_XX'
+        'XX_RT_URL_XX/REST/2.0/download/cf/XX_IMAGE_OR_BINARY_OCFV_ID_XX'
+        > file.png
+
 =head2 Paging
 
 All plural resources (such as C</tickets>) require pagination, controlled by
 the query parameters C<page> and C<per_page>.  The default page size is 20
 items, but it may be increased up to 100 (or decreased if desired).  Page
-numbers start at 1.
+numbers start at 1. The number of pages is returned, and if there is a next
+or previous page, then the URL for that page is returned in the next_page
+and prev_page variables respectively. It is up to you to store the required
+JSON to pass with the following page request.
+
+=head2 Disabled items
+
+By default, only enabled objects are returned. To include disabled objects
+you can specify C<find_disabled_rows=1> as a query parameter.
 
 =head2 Fields
 
@@ -629,13 +1065,14 @@ You can use additional fields parameters to expand child blocks, for
 example (line wrapping inserted for readability):
 
     XX_RT_URL_XX/REST/2.0/tickets
-      ?fields=Owner,Status,Created,Subject,Queue
+      ?fields=Owner,Status,Created,Subject,Queue,CustomFields
       &fields[Queue]=Name,Description
 
 Says that in the result set for tickets, the extra fields for Owner, Status,
-Created, Subject and Queue should be included. But in addition, for the Queue
-block, also include Name and Description. The results would be similar to
-this (only one ticket is displayed):
+Created, Subject, Queue and CustomFields should be included. But in
+addition, for the Queue block, also include Name and Description. The
+results would be similar to this (only one ticket is displayed in this
+example):
 
    "items" : [
       {
@@ -656,7 +1093,18 @@ this (only one ticket is displayed):
             "Name" : "General",
             "Description" : "The default queue",
             "_url" : "XX_RT_URL_XX/REST/2.0/queue/1"
-         }
+         },
+         "CustomFields" : [
+             {
+                 "id" : "1",
+                 "type" : "customfield",
+                 "_url" : "XX_RT_URL_XX/REST/2.0/customfield/1",
+                 "name" : "My Custom Field",
+                 "values" : [
+                     "CustomField value"
+                 },
+             }
+         ]
       }
       { … },
       …
@@ -805,7 +1253,7 @@ L<rt.cpan.org|http://rt.cpan.org/Public/Dist/Display.html?Name=RT-Extension-REST
 
 =head1 LICENSE AND COPYRIGHT
 
-This software is Copyright (c) 2015-2019 by Best Practical Solutions, LLC.
+This software is Copyright (c) 2015-2020 by Best Practical Solutions, LLC.
 
 This is free software, licensed under:
 
