@@ -5,7 +5,7 @@ use warnings;
 use Moose::Role;
 use namespace::autoclean;
 use JSON ();
-use RT::Extension::REST2::Util qw( deserialize_record error_as_json expand_uid );
+use RT::Extension::REST2::Util qw( deserialize_record error_as_json expand_uid update_custom_fields );
 use List::MoreUtils 'uniq';
 
 with 'RT::Extension::REST2::Resource::Role::RequestBodyIsJSON'
@@ -21,11 +21,83 @@ sub create_path {
     $_[0]->record->id || undef
 }
 
-sub content_types_accepted { [ {'application/json' => 'from_json'} ] }
+sub content_types_accepted { [ {'application/json' => 'from_json'}, { 'multipart/form-data' => 'from_multipart' } ] }
+
+sub from_multipart {
+    my $self = shift;
+    my $json_str = $self->request->parameters->{JSON};
+    return error_as_json(
+        $self->response,
+        \400, "Json is a required field for multipart/form-data")
+            unless $json_str;
+
+    my $json = JSON::decode_json($json_str);
+
+    my $cfs = delete $json->{CustomFields};
+    if ($cfs) {
+        foreach my $id (keys %$cfs) {
+            my $value = delete $cfs->{$id};
+
+            if (ref($value) eq 'ARRAY') {
+                my @values;
+                foreach my $single_value (@$value) {
+                    if ( ref $single_value eq 'HASH' && ( my $field_name = $single_value->{UploadField} ) ) {
+                        my $file = $self->request->upload($field_name);
+                        if ($file) {
+                            open my $filehandle, '<', $file->tempname;
+                            if (defined $filehandle && length $filehandle) {
+                                my ( @content, $buffer );
+                                while ( my $bytesread = read( $filehandle, $buffer, 72*57 ) ) {
+                                    push @content, MIME::Base64::encode_base64($buffer);
+                                }
+                                close $filehandle;
+
+                                push @values, {
+                                    FileName    => $file->filename,
+                                    FileType    => $file->headers->{'content-type'},
+                                    FileContent => join("\n", @content),
+                                };
+                            }
+                        }
+                    }
+                    else {
+                        push @values, $single_value;
+                    }
+                }
+                $cfs->{$id} = \@values;
+            }
+            elsif ( ref $value eq 'HASH' && ( my $field_name = $value->{UploadField} ) ) {
+                my $file = $self->request->upload($field_name);
+                if ($file) {
+                    open my $filehandle, '<', $file->tempname;
+                    if (defined $filehandle && length $filehandle) {
+                        my ( @content, $buffer );
+                        while ( my $bytesread = read( $filehandle, $buffer, 72*57 ) ) {
+                            push @content, MIME::Base64::encode_base64($buffer);
+                        }
+                        close $filehandle;
+
+                        $cfs->{$id} = {
+                            FileName    => $file->filename,
+                            FileType    => $file->headers->{'content-type'},
+                            FileContent => join("\n", @content),
+                        };
+                    }
+                }
+            }
+            else {
+                $cfs->{$id} = $value;
+            }
+        }
+        $json->{CustomFields} = $cfs;
+    }
+
+    return $self->from_json($json);
+}
 
 sub from_json {
     my $self = shift;
-    my $params = JSON::decode_json( $self->request->content );
+    my $params = shift || JSON::decode_json( $self->request->content );
 
     %$params = (
         %$params,
@@ -52,94 +124,18 @@ sub update_record {
         AttributesRef => [ $self->record->WritableAttributes ],
     );
 
-    push @results, $self->_update_custom_fields($data->{CustomFields});
+    push @results, update_custom_fields($self->record, $data->{CustomFields});
     push @results, $self->_update_role_members($data);
     push @results, $self->_update_disabled($data->{Disabled})
       unless grep { $_ eq 'Disabled' } $self->record->WritableAttributes;
+    push @results, $self->_update_privileged($data->{Privileged})
+      unless grep { $_ eq 'Privileged' } $self->record->WritableAttributes;
 
     # XXX TODO: Figure out how to return success/failure?  Core RT::Record's
     # ->Update will need to be replaced or improved.
     return @results;
 }
 
-sub _update_custom_fields {
-    my $self = shift;
-    my $data = shift;
-
-    my $record = $self->record;
-    my @results;
-
-    foreach my $cfid (keys %{ $data }) {
-        my $val = $data->{$cfid};
-
-        my $cf = $record->LoadCustomFieldByIdentifier($cfid);
-        next unless $cf->ObjectTypeFromLookupType($cf->__Value('LookupType'))->isa(ref $record);
-
-        if ($cf->SingleValue) {
-            if (ref($val) eq 'ARRAY') {
-                $val = $val->[0];
-            }
-            elsif (ref($val)) {
-                die "Invalid value type for CustomField $cfid";
-            }
-
-            my ($ok, $msg) = $record->AddCustomFieldValue(
-                Field => $cf,
-                Value => $val,
-            );
-            push @results, $msg;
-        }
-        else {
-            my %count;
-            my @vals = ref($val) eq 'ARRAY' ? @$val : $val;
-            for (@vals) {
-                $count{$_}++;
-            }
-
-            my $ocfvs = $cf->ValuesForObject( $record );
-            my %ocfv_id;
-            while (my $ocfv = $ocfvs->Next) {
-                my $content = $ocfv->Content;
-                $count{$content}--;
-                push @{ $ocfv_id{$content} }, $ocfv->Id;
-            }
-
-            # we want to provide a stable order, so first go by the order
-            # provided in the argument list, and then for any custom fields
-            # that are being removed, remove in sorted order
-            for my $key (uniq(@vals, sort keys %count)) {
-                my $count = $count{$key};
-                if ($count == 0) {
-                    # new == old, no change needed
-                }
-                elsif ($count > 0) {
-                    # new > old, need to add new
-                    while ($count-- > 0) {
-                        my ($ok, $msg) = $record->AddCustomFieldValue(
-                            Field => $cf,
-                            Value => $key,
-                        );
-                        push @results, $msg;
-                    }
-                }
-                elsif ($count < 0) {
-                    # old > new, need to remove old
-                    while ($count++ < 0) {
-                        my $id = shift @{ $ocfv_id{$key} };
-                        my ($ok, $msg) = $record->DeleteCustomFieldValue(
-                            Field   => $cf,
-                            ValueId => $id,
-                        );
-                        push @results, $msg;
-                    }
-                }
-            }
-        }
-    }
-
-    return @results;
-}
-
 sub _update_role_members {
     my $self = shift;
     my $data = shift;
@@ -166,10 +162,18 @@ sub _update_role_members {
                 die "Invalid value type for role $role";
             }
 
-            my ($ok, $msg) = $record->AddWatcher(
-                Type => $role,
-                User => $val,
-            );
+            my ($ok, $msg);
+            if ($record->can('AddWatcher')) {
+                ($ok, $msg) = $record->AddWatcher(
+                    Type => $role,
+                    User => $val,
+                );
+            } else {
+                ($ok, $msg) = $record->AddRoleMember(
+                    Type => $role,
+                    User => $val,
+                );
+            }
             push @results, $msg;
         }
         else {
@@ -266,6 +270,22 @@ sub _update_disabled {
     return @results;
 }
 
+sub _update_privileged {
+    my $self = shift;
+    my $data = shift;
+    my @results;
+
+    my $record = $self->record;
+    return unless defined $data and $data =~ /^[01]$/;
+
+    return unless $record->can('SetPrivileged');
+
+    my ($ok, $msg) = $record->SetPrivileged($data);
+    push @results, $msg;
+
+    return @results;
+}
+
 sub update_resource {
     my $self = shift;
     my $data = shift;
@@ -290,6 +310,55 @@ sub create_record {
 
     my $cfs = delete $args{CustomFields};
 
+    # Lookup CustomFields by name.
+    if ($cfs) {
+        foreach my $id (keys(%$cfs)) {
+			my $value = delete $cfs->{$id};
+            if ( ref($value) eq 'HASH' ) {
+                foreach my $field ( 'FileName', 'FileType', 'FileContent' ) {
+                    return ( 0, 0, "$field is a required field for Image/Binary ObjectCustomFieldValue" )
+                        unless $value->{$field};
+                }
+                $value->{Value}        = delete $value->{FileName};
+                $value->{ContentType}  = delete $value->{FileType};
+                $value->{LargeContent} = MIME::Base64::decode_base64( delete $value->{FileContent} );
+            }
+            elsif ( ref($value) eq 'ARRAY' ) {
+                my $i = 0;
+                foreach my $single_value (@$value) {
+                    if ( ref($single_value) eq 'HASH' ) {
+                        foreach my $field ( 'FileName', 'FileType', 'FileContent' ) {
+                            return ( 0, 0,
+                                "$field is a required field for Image/Binary ObjectCustomFieldValue" )
+                                unless $single_value->{$field};
+                        }
+                        $single_value->{Value}       = delete $single_value->{FileName};
+                        $single_value->{ContentType} = delete $single_value->{FileType};
+                        $single_value->{LargeContent}
+                            = MIME::Base64::decode_base64( delete $single_value->{FileContent} );
+                        $value->[$i] = $single_value;
+                    }
+                    $i++;
+                }
+            }
+            $cfs->{$id} = $value;
+
+            if ($id !~ /^\d+$/) {
+                my $cf = $record->LoadCustomFieldByIdentifier($id);
+
+                if ($cf->Id) {
+                    $cfs->{$cf->Id} = $cfs->{$id};
+                    delete $cfs->{$id};
+                } else {
+                    # I would really like to return an error message, but, how?
+                    # RT appears to treat missing permission to a CF or
+                    # non-existance of a CF as a non-fatal error.
+                    RT->Logger->error( $record->loc( "Custom field [_1] not found", $id ) );
+                }
+            }
+        }
+    }
+
     # if a record class handles CFs in ->Create, use it (so it doesn't generate
     # spurious transactions and interfere with default values, etc). Otherwise,
     # add OCFVs after ->Create
@@ -306,7 +375,7 @@ sub create_record {
     my ($ok, @rest) = $record->$method(%args);
 
     if ($ok && $cfs) {
-        $self->_update_custom_fields($cfs);
+        update_custom_fields($record, $cfs);
     }
 
     return ($ok, @rest);
