@@ -40,6 +40,13 @@ around 'serialize' => sub {
     my $self = shift;
     my $data = $self->$orig(@_);
     $data->{Privileged} = $self->record->Privileged ? 1 : 0;
+
+    unless ( $self->record->CurrentUserHasRight("AdminUsers") or $self->record->id == $self->current_user->id ) {
+        my $summary = { map { $_ => $data->{$_} }
+                ( '_hyperlinks', 'Privileged', split( /\s*,\s*/, RT->Config->Get('UserSummaryExtraInfo') ) ) };
+        return $summary;
+    }
+
     $data->{Disabled}   = $self->record->PrincipalObj->Disabled;
     $data->{Memberships} = [
         map { expand_uid($_->UID) }
@@ -52,20 +59,44 @@ sub forbidden {
     my $self = shift;
     return 0 if not $self->record->id;
     return 0 if $self->record->id == $self->current_user->id;
-    return 0 if $self->record->CurrentUserHasRight("AdminUsers");
+    return 0 if $self->current_user->Privileged;
     return 1;
 }
 
 sub hypermedia_links {
     my $self = shift;
     my $links = $self->_default_hypermedia_links(@_);
-    push @$links, $self->_transaction_history_link;
 
-    my $id = $self->record->id;
-    push @$links,
-      { ref  => 'memberships',
-        _url => RT::Extension::REST2->base_uri . "/user/$id/groups",
-      };
+    if (   $self->record->CurrentUserHasRight('ShowUserHistory')
+        or $self->record->CurrentUserHasRight("AdminUsers")
+        or $self->record->id == $self->current_user->id )
+    {
+        push @$links, $self->_transaction_history_link;
+    }
+
+    # TODO Use the same right in UserGroups.pm.
+    # Maybe loose it a bit so user can see groups?
+    if ((   $self->current_user->HasRight(
+                Right  => "ModifyOwnMembership",
+                Object => RT->System,
+            )
+            && $self->current_user->id == $self->record->id
+        )
+        || $self->current_user->HasRight(
+            Right  => 'AdminGroupMembership',
+            Object => RT->System
+        )
+       )
+    {
+
+        my $id = $self->record->id;
+        push @$links,
+            {
+            ref  => 'memberships',
+            _url => RT::Extension::REST2->base_uri . "/user/$id/groups",
+            };
+    }
+
     return $links;
 }
 
